/**
 * Async function mapping some `input` to an `[input, output]` tuple or a `null` if no output exists
 */
type MapFunction<In, Out> = (input: In) => Promise<[In, Out] | null>;

/**
 * Concurrently create some mapping of input to output elements.
 *
 * @param inputs collection of input elements
 * @param map_fn function mapping inputs to outputs
 * @returns ES6 Map object with inputs mapped to outputs where possible
 */
export async function map_concurrent<In, Out>(
  inputs: Iterable<In>,
  map_fn: MapFunction<In, Out>,
): Promise<Map<In, Out>> {
  // apply map_fn to inputs
  const queue = [...inputs].map((input) => map_fn(input));

  // wait for calls to finish
  return new Map<In, Out>(
    (await Promise.all(queue)).filter((job): job is [In, Out] => job !== null),
  );
}
